#! /bin/sh
# \
exec tclsh${TCL} "$0" ${1+"$@"}

#
# ThatchTool
#
# (formerly TkThatch)
#
#
# Stuart Cassoff
#
# December 2009, April 2010, Winter 2012, Summer 2012, Spring 2015, Winter 2018, Spring 2018
# Version 0.1
#
# Spring 2018
# Version 0.2
#
# Summer 2018
# Version 0.3
#
# Summer 2022
# Version 0.4
#

[namespace eval thatchtool {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	thatchtool
		title	ThatchTool
		descr	{Thatch Tool}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.4
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2009,2018 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


#---------------------------------------------------------------------
# Info
#---------------------------------------------------------------------

proc readme {} { return [string range {
<Control-q> will close windows or exit.

Home is optional.
Infile is optional.

To add files:
  Right-click in file name box and select a file.
  Erase mtime if mtime is not to be stored.
  Use checkbuttons to select whether to store size
  or enable base64 or zip transforms
  or md5 and sha256 chacksums,
  then click on the Keep button.

Thatchtool will load thatchtool project files (.tp)
from the command line.

"CSize" is Chunk Size.
Chunk Size is size of stored file + Thatch header inside the thatched file.

} 1 end-1]}

proc usage {} {
	return [format {%s%susage: %s [-?] [options] [filename] [...]} [cfg fullname] \n [string tolower [cfg name]]]
}

proc help {} {
	return [subst -nocommands [join {
		{-l  List (default)}
		{-c  Create}
		{-a  Add}
		{-r  Remove}
		{-e  Get}
		{-y  Copy}
		{-h  Home}
		{-H  Home + Thatch file home}
		{-o  fn Output file}
		{-m  fn Main}
		{-wp Wish preamble}
		{-np No preamble}
		{-pp Print preamble}
		{-pb Print bootstrap}
		{-tb Use Thatch bootstrap}
		{-g  Gui mode}
		{-v  Increase verbosity}
		{-?  This info}
	} \n]]
}

proc messages {} {
	return [join [cfg msgs] \n]
}

proc resetMessages {} {
	cfg msgs {}
}

proc log {msg {status 1}} {
	if {[catch {
		cfg+ msgs $msg
		if {[cfg guiCreated]} {
			if {$status} {
				setStatus $msg
			}
		} else {
			puts [cfg stdout] $msg
		}
	}]} {
		#puts stderr $msg
	}
}

#---------------------------------------------------------------------
# Misc
#---------------------------------------------------------------------

proc filterNv {data flds} {
	set res {}
	foreach l $data { if {[lindex $l 0] in $flds} { lappend res {*}$l } }
	return $res
}

proc mapCalc {ratio offset size width} {
	set res {}
	set sp [expr {int($offset * $ratio)}]
	set sx [expr {$sp % $width}]
	set sy [expr {$sp / $width}]
	set ep [expr {int(($offset + $size) * $ratio)}]
	set ex [expr {$ep % $width}]
	set ey [expr {$ep / $width}]

	if {$ey == $sy} {
		lappend res [list $sx $sy $ex $ey] {} {}
	} else {
		if {$sx == 0} {
			lappend res {}
		} else {
			lappend res [list $sx $sy [expr {$width + 1}] $sy]
			incr sy
		}
		if {$ey > $sy} {
			lappend res [list 0 $sy [expr {$width + 1}] $ey]
		} else {
			lappend res {}
		}
		if {$ex == $width} {
			lappend res {}
		} else {
			lappend res [list 0 $ey [expr {$ex + 1}] $ey]
		}
	}

	return $res
}

proc hexDump {d {w 8}} {
	set l [string length $d]
	set z {}
	for {set i 0; set ii [expr {$w - 1}]; set addr 0} {$i < $l} {incr i $w; incr ii $w; incr addr $w} {
		binary scan [string range $d $i $ii] @0H*@0a* hex ascii
		regsub -all -- {[^[:graph:] ]} $ascii {.} ascii
		regsub -all -- {..} $hex {& } hex
		lappend z [format "%04X  %-[expr {3 * $w}]s %-${w}s" $addr $hex $ascii]
	}
	return $z
}

proc moveFile {src dst} {
	if {1} {
	file rename -force $src $dst
	} else {
	file copy -force $src $dst
	file delete -force $src
	}
}

#---------------------------------------------------------------------
# Misc GUI
#---------------------------------------------------------------------

proc adjMainGeom {} {
	pjs!
	if {[dict size $pjs] == 0} {
		wm geometry . 300x100
	} elseif {[dict size $pjs] == 1} {
		wm geometry . {}
#		update idletasks
	}
	after idle [list [cfg NS]::centreToplevel .]
}

proc colTextWidth {tv txt} {
	return [expr {[font measure [ttk::style lookup [$tv cget -style] -font] $txt] + 30}]
}

proc adjTv {tv {cols {}}} {
	if {[llength $cols] == 0} {
		set cols [$tv columns]
	}
	foreach col $cols {
		set width [colTextWidth $tv [$tv heading $col -text]]
		foreach item [$tv children {}] {
			set txtwidth [colTextWidth $tv [$tv set $item $col]]
			if {$txtwidth > $width} {
				set width $txtwidth
			}
		}
		$tv column $col -width $width -minwidth $width
	}
}

proc autoscrollbar {w args} {
	set w [ttk::scrollbar $w {*}$args]
	::autoscroll::autoscroll $w
	return $w
}

proc changeWordWrap {} {
	set w [epmW]
	$w configure -wrap [expr {[$w cget -wrap] eq "none" ? "word" : "none"}]
}

#---------------------------------------------------------------------
# Dialog
#---------------------------------------------------------------------

proc oc {title message} {
	set choice [dialog [randWindow] $title $message "" 1 "&Ok" &Cancel"]
	if {$choice == -1} { set choice end }
	return [string index OC $choice]
}

proc yn {title message} {
	set choice [dialog [randWindow] $title $message "" 1 "&Yes" "&No"]
	if {$choice == -1} { set choice end }
	return [string index YN $choice]
}

proc okcancel {title message} {
	return [expr {[oc $title $message] eq "O"}]
}

proc yesno {title message} {
	return [expr {[yn $title $message] eq "Y"}]
}

proc dialog {window title text bitmap default args} {
	set ns [cfg NS]:
	destroyToplevel $window
	newToplevel $window
	groupToplevel $window [gui W]
	setToplevelTitle $window $title
	transientizeToplevel $window [gui W]
	set t [split $text \n]
	set weegs [dict create]
	eval [gub "
		[gubMultiGob [llength $t] L F]
		.{L} -text $t
		.{B} -command [lmap i [::gub::ExpandRange 0-[expr {[llength $args]-1}]] { list $ns:dialogClick $i}]
		;F [string repeat {B } [llength $args]]
		[gubButtonGob -buttonchar B -modifiers Alt {*}$args]
		[expr {$default > -1 ? "/B" : ""}]
	" $window weegs]
	foreach k {Control-q Control-Q Escape} { bind $window <$k> [list $ns:dialogClick -1] }
	onToplevelDelete $window [list $ns:dialogClick -1]
	centreToplevel $window [gui W]
#	wm attributes $window -topmost 1
	grab set $window
	if {$default > -1} { focus [dict get $weegs B$default] }
	gui dialogRetval -1
	vwait [gui& dialogRetval]
	grab release $window
	destroy $window
	return [gui dialogRetval]
}

proc dialogClick {what} {
	gui dialogRetval $what
}

#---------------------------------------------------------------------
# Messages
#---------------------------------------------------------------------

proc clearMessages {t} {
	$t delete 1.0 end
	resetMessages
}

proc viewMessages {} {
	set w .q
	destroy $w
	set w [newToplevel $w]
	groupToplevel $w [gui W]
	set weegs [dict create]
	eval [gub "
		;F t S
		;  C
		;F B B
		.B0 -command {[list after idle [list destroy $w]]}
		.B1 -command {[cfg NS]::clearMessages /t0/}
		:F0,t0
		-.,F0 ; |.,F0
		/t0 viewText
		/B0 okButton
		[gubButtonGob -buttonchar B &Ok &Clear]
		&B0 Control+q
		&B0 Escape
#		/. W
	" $w weegs]
	setToplevelTitle $w [string cat [getToplevelTitle [gui W]] " - Messages"]
	[dict get $weegs viewText] insert end [messages]
	focus [dict get $weegs okButton]
	centreToplevel $w [gui W]
}

#---------------------------------------------------------------------
# Config
#---------------------------------------------------------------------

proc configAdj {} {
	if {[cfg theme] eq ""} { cfg theme default }
	#set cfg(configFile) [file normalize $cfg(configFile)]
}

proc configLoad {{fn {?}}} {
	if {$fn eq {}} { set fn [cfg configFile] }
	if {![lassign [loadListFile $fn -filetypes [list [dict get [cfg filetypes] config]]] fn data]} { return 0 }
	dict for {k v} [filterNv $data [cfg configFileFields]] { cfg $k $v }
	cfg configFile $fn
	configAdj
	return 1
}

proc configSave {{fn {}}} {
	if {$fn eq {}} { set fn [cfg configFile] }
	foreach n [cfg configFileFields] { lappend l [list $n [cfg $n]] }
	if {![lassign [saveFile $fn [join $l \n]\n [list [dict get [cfg filetypes] config]]] fn]} { return 0 }
	cfg configFile $fn
	configAdj
	return 1
}

proc autoloadConfig {} {
	configAdj
	#if {[info exists cfg(pConfigFile)]} { set cfg(configFile) $cfg(pConfigFile) }
	configAdj
	if {[cfg autoloadConfigFile]} { configLoad {} }
	configAdj
}

proc cfgThemeChanged {w} {
	ttk::style theme use [$w get]
}

proc configGo {} {
	::aitcs::aitcs new ecfg
	foreach f [cfg configBoolFields] { ecfg $f [cfg $f] }
	set wugs [createConfigGui]
	cfg->gui $wugs
	grab set [dict get $wugs W]
	focus [dict get $wugs B6]
}

proc configStop {wugs} {
	ttk::style theme use [cfg theme]
	grab release [dict get $wugs W]
	after idle [list destroy [dict get $wugs W]]
	ecfg---
}

proc configSaveStop {wugs} {
	foreach f [cfg configBoolFields] { cfg $f [ecfg $f] }
	gui->cfg $wugs
	configSave
	configStop $wugs
}

proc cfg->gui {wugs} {
	foreach p [cfg autoloadableProjects] {
		[dict get $wugs autoloadableProjects] insert {} end -values $p
	}
	[dict get $wugs themeSelector] current [lsearch [ttk::style theme names] [cfg theme]]
}

proc gui->cfg {wugs} {
	cfg autoloadableProjects {}
	foreach item [[dict get $wugs autoloadableProjects] children {}] {
		cfg+ autoloadableProjects [dict values [[dict get $wugs autoloadableProjects] set $item]]
	}
	cfg theme [[dict get $wugs themeSelector] get]
}

proc configAddProject {wugs} {
	set fn [tk_getOpenFile -filetypes [list [dict get [cfg filetypes] tp]] -parent [dict get $wugs W]]
	if {$fn eq ""} { return }
	[dict get $wugs autoloadableProjects] insert {} end -values [list $fn on]
}

proc configRemoveProject {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	[dict get $wugs autoloadableProjects] delete $item
}

proc configMoveProjectUp {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	set index [[dict get $wugs autoloadableProjects] index $item]
	if {$index == 0} { return }
	[dict get $wugs autoloadableProjects] move $item {} [incr index -1]
}

proc configMoveProjectDown {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	set index [[dict get $wugs autoloadableProjects] index $item]
	if {$index == [llength [[dict get $wugs autoloadableProjects] children {}]]} { return }
	[dict get $wugs autoloadableProjects] move $item {} [incr index]
}

proc configToggleProjectAutoload {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	[dict get $wugs autoloadableProjects] set $item autoload [expr {[[dict get $wugs autoloadableProjects] set $item autoload] ? "off" : "on"}]
}

proc createConfigGui {} {
	set w [newToplevel]
	groupToplevel $w [gui W]
	setToplevelTitle $w Configure
	set alpcols {project autoload}
	set weegs [dict create]
	eval [gub "
		;M Vp
		;  FpB B B B B
		;M LaKaLtXt
		;M KsKmKbKzKdKh
		;F B B
		.{M} -text {Autoloadable Projects} Options {Add Files Defaults}
		.Vp -columns {$alpcols} -show headings
		.{La,Lt} -text {Autoload Projects} Theme
		.Xt -state readonly -values {[ttk::style theme names]}
		.{K<smbzdh>} -text Size Mtime Base64 Zip Md5 Sha256
		.{K<asmbzdh>} -variable [lmap f [cfg configBoolFields] { ecfg& $f }]
		:M,V
		:L,K -sticky w
		:X -sticky ew -padx {4 0}
		-.,M0 ; |.,M0
		|M1 1
		[gubButtonGob -buttonchar B -modifiers Alt \
			&Add &Remove {Move &Up} {Move &Down} {&Toggle Autoload} &OK &Cancel \
		]
		/Vp autoloadableProjects
		/Xt themeSelector
		/B
		/. W
	" $w weegs]
	set tv [dict get $weegs autoloadableProjects]
	foreach col $alpcols {
		$tv heading $col -text [string totitle $col]
	}
	$tv column autoload -anchor center -stretch 0
	adjTv $tv $alpcols
	foreach b {0 1 2 3 4 5 6} cmd {configAddProject configRemoveProject configMoveProjectUp configMoveProjectDown configToggleProjectAutoload configSaveStop configStop} {
		[dict get $weegs B$b] configure -command [list [cfg NS]::$cmd $weegs]
	}
	bind [dict get $weegs themeSelector] <<ComboboxSelected>> [list [cfg NS]::cfgThemeChanged [dict get $weegs themeSelector]]
	foreach k {Control-q Control-Q Escape} { bind $w <$k> [list [cfg NS]::configStop $weegs] }
	onToplevelDelete $w [list [cfg NS]::configStop $weegs]
	centreToplevel $w [gui W]
	return $weegs
}

#---------------------------------------------------------------------
# Gui
#---------------------------------------------------------------------

proc createGui {{W {.}}} {
	pjs!;gui!
	set ns [cfg NS]:
	gui w [string trimright [gui W $W] .]
	gubWidgetMap [dict merge [gubWidgetMap] [dict create C [dict merge [dict get [gubWidgetMap] S] [dict create command $ns:autoscrollbar]]]]
	gui M [createMainMenu [gui w]]
	eval [gub -a "N ; : ; -. ; |. ; /N0 mainBook; &. Control+q $ns:fileExit" [gui w]]
	onToplevelDelete [gui W] $ns:fileExit
	catch {ttk::style theme use [cfg theme]}
	wm title [gui W] [cfg name]\ [cfg ver]
	[gui W] configure -menu [gui M]
	gui CVM [createChangeViewMenu [gui w]]
	epmCreateMenu [gui w]
	adjMainGeom
	#foreach i {3 4} { $gui(mainMenu,File) entryconfigure $i -state disabled }
}

proc createMainMenu {w} {
	set M $w.mMain
	set mob "
		.&File
		. &Configure	-command configGo
#		. &Messages	-command viewMessages
		. &Program Info	-command createProgramInfoGui
		. E&xit
		.&Project
		. &Open
		. &Save
		. Save &As	-command {[cfg NS]::projectSave ?}
		. &Close
		. &New
		.&Thatch
		. &Open		-command openThatch
	"
	eval [dict get [meb -menuname $M -namespace [cfg NS] $mob] tcl src]
	return $M
}

proc createChangeViewMenu {w} {
	set M $w.mChangeView
	set mob "
		.&Word Wrap -variable ::ww -command changeWordWrap
	"
	eval [dict get [meb -menuname $M -namespace [cfg NS] $mob] tcl src]
	return $M
}

proc createProgramInfoGui {} {
	set w .pinfo
	destroyToplevel $w
	set w [newToplevel $w]
	groupToplevel $w [gui W]
	setToplevelTitle $w "[cfg fullname] - Info"
	onToplevelDelete $w [list set ::ww $::ww]
	set weegs [dict create]
	eval [gub "
		;N FrtrC
		;    C
		;  FhthC
		;    C
		;  Fa
		;B
		[gubButtonGob -buttonchar B -modifiers Alt &Ok]
		.t -width 76 -height 20 -wrap none
		.B0 -command {after idle destroy /./}
		+{F<rha>} -text &Readme &Help &About
		:N0,t
		|.,F<rha> ; -.,F<rha>
#		&N0 <<NotebookTabChanged>> [cfg NS]::programInfoAdjustChangeViewMenuForText %W
		&B0 Control+q
		&B0 Escape
		/N0 book
		/B0 okbutton
		/Fa about
		/t<rh>
	" $w weegs]
	::ttk::notebook::enableTraversal [dict get $weegs book]
	setTextText [dict get $weegs tr] [readme]
	setTextText [dict get $weegs th] [usage]\n[help]
	::central::extras::CentralPackageBrowser new [dict get $weegs about]
	foreach t [list tr th] {
		[set t [dict get $weegs $t]] configure -state disabled 
		epmBindMenuTo $t [gui CVM]
	}
	focus [dict get $weegs okbutton]
	centreToplevel $w [gui W]
}

proc programInfoAdjustChangeViewMenuForText {w} {
	set ::ww [expr {[[$w select].t[string index [$w select] end] cget -wrap] eq "word"}]
}

#---------------------------------------------------------------------
# Project / Thatch
#---------------------------------------------------------------------

proc autoloadProjects {} {
	if {![cfg autoloadProjects]} { return }
	foreach p [cfg autoloadableProjects] {
		if {[lindex $p 1]} {
			projectOpen [lindex $p 0]
		}
	}
}

proc newProject {} {
	pjs!
	for {set i 0} {[dict exists $pjs [set id pj#$i]]} {incr i} {}
	dict set pjs $id [dict merge [cfg projectFields] [dict create id $id projectFile [string map {# ""} $id].tp et {} view {}]]
	return $id
}

proc addProjectToGUI {id} {
	pjs!
	adjMainGeom
	createProjectGUI [gui mainBook] $id
	[gui mainBook] add [set w [dict get $pjs $id w]] -text [dict get $pjs $id name]
	[gui mainBook] select $w
	return $id
}

proc projectOpen {{fn ""}} {
	pjs!
	if {![lassign [loadListFile $fn -filetypes [list [dict get [cfg filetypes] tp]]] fn data]} { return 0 }
	set id [newProject]
	dict set pjs $id [dict merge [dict get $pjs $id] [filterNv $data [dict keys [cfg projectFields]]]]
	dict set pjs $id projectFile $fn
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc projectSave {{fn {}}} {
	pjs!
	set id [getPjIdOfCurrentTab]
	gui->project $id
	if {$fn eq {}} {
		set fn [dict get $pjs $id projectFile]
	}
	if {$fn eq {}} {
		set fn ?
	}
	set l {}
	foreach f [cfg projectEditFields] {
		lappend l [list $f [dict get $pjs $id $f]]
	}
	if {![lassign [saveFile $fn [join $l \n]\n -filetypes [list [dict get [cfg filetypes] tp]]] fn]} { return 0 }
	dict set pjs $id projectFile $fn
	project->gui $id
	return 1
}

proc projectClose {} {
	pjs!
	set id [getPjIdOfCurrentTab]
	[gui mainBook] forget [dict get $pjs $id w]
	destroy [dict get $pjs $id w]
	dict unset pjs $id
	adjMainGeom
}

proc openThatch {{fn ""}} {
	pjs!
	if {$fn eq {} || $fn eq {?}} {
		gui!
		lappend filetypes [dict get [cfg filetypes] all]
		set fn [tk_getOpenFile -filetypes $filetypes -parent $gui(W) -title "Open Thatch"]
	}
	if {$fn eq {}} { return }
	set id [newProject]
	dict set pjs $id name $id
	dict set pjs $id outFile $fn
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc adjThatch {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {[file exists $fn]} {
		dict set pjs $id outFileSize [file size $fn]
		dict set pjs $id canvMapRatio [expr {(double([cfg mapCanvWidth] * [cfg mapCanvHeight])) / double([dict get $pjs $id outFileSize])}]
	} else {
		dict set pjs $id outFileSize 0
		dict set pjs $id canvMapRatio 0
	}
}

proc adjLineForDisp {id l} {
	pjs!
	dict set pjs $id mtimeFormat "%Y/%m/%d %T"
	if {[set f [dict get $pjs $id mtimeFormat]] ne ""} {
		set i [lsearch [cfg thatchTvFields] mtime]
		if {[set time [lindex $l $i]] ne ""} {
			lset l $i [clock format $time -format $f]
		}
	}
	return $l
}

proc project->gui {id} {
	pjs!
	foreach f [cfg projectEditFields] {
		[dict get $pjs $id w_$f] configure -validate none
		[dict get $pjs $id w_$f] delete 0 end
		[dict get $pjs $id w_$f] insert end [dict get $pjs $id $f]
		[dict get $pjs $id w_$f] configure -validate key
	}
	setProjectTitle $id [dict get $pjs $id name]
}

proc setProjectTitle {id t} {
	pjs!
	[gui mainBook] tab [dict get $pjs $id w] -text $t
}

proc gui->project {id} {
	pjs!
	foreach f [cfg projectEditFields] {
		dict set pjs $id $f [[dict get $pjs $id w_$f] get]
	}
	setProjectTitle $id [dict get $pjs $id name]
}

proc thatch->gui {id {focus {}}} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn] } {return }
	set offset [dict get $pjs $id outFileSize]
	set tv [dict get $pjs $id w_tv]
	set c [dict get $pjs $id w_c]
	set tf [lrange [cfg thatchTvFields] 1 end-1]
	set pos [file size $fn]
	set thIndex [::thatch::thatchFullFileGlob $fn]
	dict set pjs $id thIndex $thIndex
	$tv delete [$tv children {}]
	$c delete all
	[dict get $pjs $id w_ename] delete 0 end
	[dict get $pjs $id w_t] delete 1.0 end
	defaults->editThatchedFileInfoGui $id
	dict set pjs $id fishMap [dict create]
	foreach {v k} [lreverse [dict get $pjs $id thIndex]] {
		set l [list $k]
		if {[dict exists $v checksums]} {
			set x [dict get $v checksums]
			set r {}
			foreach {q z} [split $x ,:] {
				lappend r $q
			}
			dict set v checksums $r
		}
		foreach f [list size mtime transforms checksums offset] {
			lappend l [dgwd "" $v $f]
		}
		set csize [expr {$pos - [dict get $pjs $id thIndex $k offset]}]
		dict set pjs $id $fn $k csize $csize
		incr pos -$csize
		lappend l $csize
		dict set pjs $id fishMap [$tv insert {} 0 -values [adjLineForDisp $id $l]] $k
		createMapRegion $id $fn $k $c
	}
	$c itemconfigure all -fill [cfg mapCanvBackground]
	adjTv $tv [cfg thatchTvFields]
	if {$focus ne {}} {
		$tv focus $focus
	}
}

proc thatchedFile->gui {w id {view {}}} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn]} { return }
	dict set pjs $id view $view
	set sel [lindex [$w selection] 0]
	set name [dict get [dict get $pjs $id thIndex [dict get $pjs $id fishMap $sel]] name]
	set ext [file extension $name]
	set t [dict get $pjs $id w_t]
	set c [dict get $pjs $id w_c]
	thatchedFileInfo->editThatchedFileInfoGui $id
	set d [thatchExtract $fn [dict create $name [dict get $pjs $id thIndex $name]]]
	$t configure -state normal
	image delete {*}[$t image names]
	$t delete 1.0 end
	if {$ext in {".gif" ".png" ".ppm" ".xbm"}} {
		if {$view eq "hex"} {
			$t insert end [join [hexdump $d 32] \n]
			set title Img
		} else {
			$t image create end -image [image create [expr {$ext eq ".xbm" ? "bitmap" : "photo"}] -data $d]
			set title Hex
		}
	} else {
		set xfrms [dgwd {} $pjs $id thIndex $name transforms]
		if {"base64" in $xfrms} {
			if {$view eq {}} {
				set d [join [hexdump $d 32] \n]
			}
			set title [expr {$view eq {} ? "Txt" : "Hex"}]
		} else {
			if {$view eq "hex"} {
				set d [join [hexdump $d 32] \n]
			}
			set title [expr {$view eq {} ? "Hex" : "Txt"}]
		}
		$t insert end $d
	}
	$c itemconfigure all -fill [cfg mapCanvBackground]
	foreach item [$w selection] {
		set tag [dict get $pjs $id thIndex [dict get $pjs $id fishMap $item] name]
		$c itemconfigure $tag -fill [cfg mapCanvForeground]
		$c raise $tag
	}
	$t configure -state disabled
	[dict get $pjs $id w_bchgvw] configure -text [string cat \u2193 $title]
}

proc thatchedFileInfo->editThatchedFileInfoGui {id} {
	pjs!
	set w [dict get $pjs $id w_tv]
	set fi [$w set [lindex [$w selection] 0]]
	foreach tf [list name mtime] gf [list w_ename w_emtime] {
		[dict get $pjs $id $gf] delete 0 end
		[dict get $pjs $id $gf] insert end [dict get $fi $tf]
	}
	set x [dict get $fi transforms]
	foreach tf [list base64 zip] gf [list w_ebase64 w_ezip] {
		[dict get $pjs $id $gf] state [expr {$tf in $x ? "" : "!"}]selected
	}
	set x [dict get $fi checksums]
	foreach tf [list md5 sha256] gf [list w_emd5 w_esha256] {
		[dict get $pjs $id $gf] state [expr {$tf in $x ? "" : "!"}]selected
	}
	foreach tf [list size] gf [list w_esize] {
		set s [expr {[dict get $fi $tf] eq "" || [dict get $fi offset]  == 0 ? "!" : ""}]
		[dict get $pjs $id $gf] state ${s}selected
	}
}

proc defaults->editThatchedFileInfoGui {id} {
	pjs!
	foreach cf [list addDefSize addDefBase64 addDefZip addDefMd5 addDefSha256] gf [list w_esize w_ebase64 w_ezip w_emd5 w_esha256] {
		[dict get $pjs $id $gf] state [expr {[cfg $cf]?"":"!"}]selected
	}
	if {![cfg addDefMtime]} {
		[dict get $pjs $id w_emtime] delete 0 end
	}
}

proc getEditThatchedFileInfoGuiValues {id} {
	pjs!
	set d [dict create]
	foreach f [list name mtime] gf [list w_ename w_emtime] {
		dict set d $f [[dict get $pjs $id $gf] get]
	}
	foreach tf [list size base64 zip md5 sha256] gf [list w_esize w_ebase64 w_ezip w_emd5 w_esha256] {
		dict set d $tf [expr {"selected" in [[dict get $pjs $id $gf] state]}]
	}
	return $d
}

proc projectNew {} {
	pjs!
	set id [newProject]
	dict set pjs $id name $id
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc getPjIdOfCurrentTab {} {
	pjs!
	set w [[gui mainBook] select]
	dict for {id pj} $pjs {
		if {[dict get $pj w] eq $w} {
			return $id
		}
	}
	return {}
}

proc woif {id} {
	pjs!
	dict set pjs $id et {}
	gui->project $id
}
proc boing {P} {
	pjs!
	set id [getPjIdOfCurrentTab]
	if {[dict get $pjs $id et] eq {}} {
		dict set pjs $id et [after idle [list [cfg NS]::woif $id]]
	}
	return true
}
proc createProjectGUI {W id} {
	pjs!
	set ns [cfg NS]:
	set flds [cfg projectEditFields]
	set w [ttk::frame [randWindow $W.fF_]]
	set gob "
		;F c L E
		[string repeat ";    L E\n" [expr {[llength $flds]-1}]]
		;P F V C
		;    C
		;  F F L E L E K K K K K B B B Bv
		;    F t C
		;      C
		.c0 -width [cfg mapCanvWidth] -height [cfg mapCanvHeight] -background [cfg mapCanvBackground] -highlightthickness 0
		.{L} -text [lmap f $flds {string totitle $f}]
		.{L[expr {[llength $flds]}],[expr {[llength $flds]+1}]} -text Name Mtime
		.E0-[expr {[llength $flds]-1}] -validate key -validatecommand {$ns:boing %P}
		.{K} -text Size Base64 Zip Md5 Sha256
		.V0 -columns {[cfg thatchTvFields]} -show headings
		[gubButtonGob -buttonchar B -modifiers Alt Keep Remove Def \u2193Hex]
		.B0 -command {$ns:alterThatch $id}
		.B1 -command {$ns:removeFileFromThatch $id}
		.B2 -command {$ns:defaults->editThatchedFileInfoGui $id}
		.Bv -command {$ns:changeView $id} -width -5
		.B0-2 -width 0
		.t0 -height 10 -state disabled -wrap none
		:E -sticky ew
		:L -sticky w
		:c0 -rowspan [llength $flds]
		:F0,3,4,P0,V0,t0
		|.,F1,2,4;-F1,4
		|F0 2
		|F3 1
		-.,F0,2 1
		/c0 $id w_c
		/{E} {} [lmap f $flds { string cat $id " " w_ $f } ] {$id w_ename} {$id w_emtime}
		/V0 $id w_tv
		/t0 $id w_t
		/{K} {} {$id w_esize} {$id w_ebase64} {$id w_ezip} {$id w_emd5} {$id w_esha256}
		/Bv $id w_bchgvw
		/. {$id w}
	"
	eval [gub $gob $w pjs]

	set tv [dict get $pjs $id w_tv]
	foreach col [cfg thatchTvFields] {
		$tv heading $col -text [string totitle $col]
		$tv column $col -stretch 0 -anchor e
	}
	$tv column [lindex [cfg thatchTvFields] 0] -stretch 1 -anchor w
	$tv column [lindex [cfg thatchTvFields] end-1] -stretch 0 -anchor w
	adjTv $tv [cfg thatchTvFields]
	bind $tv <<TreeviewSelect>> [list $ns:thatchedFile->gui %W $id]
	foreach f [concat $flds [list ename emtime]] {
		epmBindMenuTo [dict get $pjs $id w_$f] [epmMenu]
	}
	foreach f [list esize ebase64 ezip emd5 esha256] {
		[dict get $pjs $id w_$f] state !alternate
	}
	epmBindMenuTo [dict get $pjs $id w_t] [gui CVM]

	return $id
}

proc createMapRegion {id fn k c} {
	pjs!
	lassign [lrepeat 3 [list]] topLine midRect botLine
	set csize [dict get $pjs $id $fn $k csize]
	incr csize -1
	while {[llength $topLine] < 1 && [llength $midRect] < 1 && [llength $botLine] < 1} {
		lassign [mapCalc [dict get $pjs $id canvMapRatio] [dict get $pjs $id thIndex $k offset] [incr csize] [cfg mapCanvWidth]] topLine midRect botLine
	}
	if {[llength $topLine] > 0} { $c create line $topLine -tag $k }
	if {[llength $botLine] > 0} { $c create line $botLine -tag $k }
	if {[llength $midRect] > 0} { $c create rectangle $midRect -outline {} -tag $k }
}

proc changeView {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn]} { return }
	set w [dict get $pjs $id w_tv]
	set sel [lindex [$w selection] 0]
	if {[llength $sel] == 0} { return }
	set name [dict get [dict get $pjs $id thIndex [dict get $pjs $id fishMap $sel]] name]
	set view [dict get $pjs $id view]
	if {[file extension $name] in {".gif" ".png" ".ppm" ".xbm"}} {
		if {$view eq {}} {
			set view hex
		} else {
			set view {}
		}
		thatchedFile->gui $w $id $view
	} elseif {[file extension $name] in {".tcl" ".tm" ".txt" ".man" ".n" ".html" ".terms"}} {
		if {$view eq {}} {
			set view hex
		} else {
			set view {}
		}
		thatchedFile->gui $w $id $view
	}
}

#---------------------------------------------------------------------
# Thatch manip
#---------------------------------------------------------------------

proc alterThatch {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {$fn eq ""} { return }
	set gd [getEditThatchedFileInfoGuiValues $id]
	set name [dict get $gd name]
	if {$name eq ""} { return }
	set edit [dict exists $pjs $id thIndex $name]
	set ok 0
	set opts {}
	if {$edit} {
		set fid [dict get $pjs $id thIndex $name]
		foreach tf [list size] {
			set gv [dict get $gd $tf]
			set ke [dict exists $fid $tf]
			if {$ke == $gv} { continue }
			lappend opts -[expr {$gv ? "" : "no"}]$tf
		}
		foreach tf [list transforms checksums] stfl [list [list base64 zip] [list md5 sha256]] {
			set v [split [dgwd {} $fid $tf]]
			foreach f $stfl {
				set gv [dict get $gd $f]
				set ke [expr {$f in $v}]
				if {$ke != $gv} {
					lappend opts -[expr {$gv ? "" : "no"}]$f
				}
			}
		}
		file tempfile tmpfile
		set ok [thatchCopy $fn $tmpfile {*}[join $opts] $name]
	} else {
		foreach tf [list size base64 zip md5 sha256] {
			set gv [dict get $gd $tf]
			lappend opts -[expr {$gv ? "" : "no"}]$tf
		}
		file tempfile tmpfile
		set ok [thatchAdd $fn $tmpfile {*}[join $opts] $name]
	}
	if {$ok} {
		moveFile $tmpfile $fn
		adjThatch $id
		thatch->gui $id
	}
}

proc removeFileFromThatch {id} {
	pjs!
	set w [dict get $pjs $id w_tv]
	set sel [$w selection]
	if {[llength $sel] == 0} { return }
	set S [expr {[llength $sel] > 1 ? "s" : ""}]
	set msg {}
	set copyArgs {}
	foreach s $sel {
		set name [dict get [$w set $s] name]
		lappend msg $name
		lappend copyArgs -remove $name
	}
	if {![yesno "Remove File$S" [join [linsert $msg 0 "Remove File$S?"] \n]]} { return }
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	file tempfile tmpfile
	if {![thatchCopy $fn $tmpfile {*}[join $copyArgs]]} { return }
	moveFile $tmpfile $fn
	adjThatch $id
	thatch->gui $id
}

proc preamble {} {
	return [::thatch::preamble [cfg wishpreamble]]
}

proc boot {} {
	return [string map [list @VER@ [package provide thatch] @THATCHOPTS@ [cfg thatchbootopts] @MAIN@ [cfg main]] $::thatch::bootstrap]
}

proc cgogo {} {
	switch -exact [cfg action] {
	list {
		set d [thatch::thatchGlob [cfg inFile]]
		if {[cfg verbosity] > 1} {
			dict for {k v} $d {
				puts stdout "$k $v"
			}
		} else {
			dict for {k v} $d {
				puts stdout $k
			}
		}
	}
	create {
		if {[cfg thatchboot]} {
			cfg cliFiles [list [cfg inFile] {*}[cfg cliFiles]]
			lassign [chan pipe] r w
			cfg inFile @$r
			if {[cfg preamble]} {
				puts $w [preamble]
			}
			puts $w [boot]
			close $w
		}
		thatch::thatch[string totitle [cfg action]] [cfg inFile] [cfg outFile] {*}[cfg cliFiles]
	}
	add - copy {
		thatch::thatch[string totitle [cfg action]] [cfg inFile] [cfg outFile] {*}[cfg cliFiles]
	}
	get {
#puts " [cfg inFile] [cfg cliFiles]"
		set res {}
		dict for {k v} [::thatch::thatchGlob [cfg inFile] [cfg cliFiles]] {
			lappend res [thatch::thatchExtract [cfg inFile] [dict create $k $v]]
		}
		puts stdout [join $res \n]
	}
	}

}

#---------------------------------------------------------------------
# Setup / Startup / Quit
#---------------------------------------------------------------------

proc fileExit {} {
	#exit
	destroy .
}

proc args {args} {
	set c [dict create]
	if {[llength $args] == 0} { dict set c usage 1; return $c }
	dict set c verbosity [cfg verbosity]
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-l - -c - -a - -r - -e - -y {
				if {[dict exists $c action]} {
					puts stderr "options -lcarey are mutually exclusive"; exit 1
				}
				dict set c action [string map [list -l list -c create -a add -r remove -e get -y copy] $arg]
			}
			-tb { dict set c thatchboot 1 }
			-wp { dict set c wishpreamble 1 }
			-np { dict set c preamble 0 }
			-g  { dict set c gui 1 }
			-v  { dict incr c verbosity }
			-pp { dict set c printpreamble 1 }
			-pb { dict set c printboot 1 }
			-o - -h - -H - -m - -to { set state $arg }
			-?  { dict set c help 1 }
			default {
				if {![dict exists $c inFile]} {
					dict set c inFile $arg
				} else {
					dict lappend c cliFiles $arg
				}
			}
			}
		}
		-o  { set state opt; dict set c outFile $arg }
		-h  { set state opt; set home $arg }
		-H  { set state opt; set Home $arg }
		-m  { set state opt; dict set c main $arg }
		-to { set state opt; dict set c thatchbootopts $arg }
		}; #switch
	}
	if {[dict exists $c action] && [dict exists $c gui] && [dict get $c gui]} {
		puts stderr "options -lcarey and -g are mutually exclusive"; exit 1
	}
	if {[info exists Home]} {
		set home $Home
		if {![dict exists $c cliFiles]} { dict set c cliFiles [list] }
		dict set c cliFiles [linsert [dict get $c cliFiles] 0 -home $home]
	}
	if {[info exists home]} {
		dict set c inFile [file join $home [dict get $c inFile]]
	}
	return $c
}

proc setup {} {
	::aitcs::aitcs+ new cfg
	eval cfg [join [lrange [split {
		NS			[namespace current]
		name			ThatchTool
		ver			0.4
		author			"Stuart Cassoff"
		when			"Summer 2022"
		msgs			{}
		projectFields		[dict create
						name        ""
						projectFile ""
						home        ""
						inFile      ""
						outFile     ""
						lastFileDir ""
						mtimeFormat ""
					]
		projectEditFields	[list
						name
						projectFile
						home
						inFile
						outFile
					]
		thatchFields		[list
						name
						size
						mtime
						base64
						zip
						md5
						sha256
						offset
						csize
					]
		thatchTvFields		[list
						name
						size
						mtime
						transforms
						checksums
						offset
						csize
					]
		mapCanvWidth		40
		mapCanvHeight		60
		mapCanvForeground	pink
		mapCanvBackground	darkgreen
		theme			default
		autoloadableProjects	[list]
		autoloadProjects	0
		addDefSize		0
		addDefMtime		0
		addDefBase64		0
		addDefZip		0
		addDefMd5		0
		addDefSha256		0
		configFile		thatchtool.conf
		autoloadConfigFile	1
		configFileFields	[list
						theme
						autoloadableProjects
						autoloadProjects
						addDefSize
						addDefMtime
						addDefBase64
						addDefZip
						addDefMd5
						addDefSha256
					]
		configBoolFields	[list
						autoloadProjects
						addDefSize
						addDefMtime
						addDefBase64
						addDefZip
						addDefMd5
						addDefSha256
					]
		filetypes 		[dict create
						tp	{{Thatch Project Files} *.tp}
						config	{{Thatch Configuration Files} .conf}
						all	{{All Files} *}
					]
		gui			no
		usage			0
		help			0
		verbosity		1
		action			list
		inFile			""
		outFile			"-"
		cliFiles		{}
		printpreamble		0
		printboot		0
		main			""
		thatchboot		0
		thatchbootopts		""
		wishpreamble		0
		preamble		1
	} \n] 1 end-1] " \\\n"]

	cfg fullname [cfg name]\ [cfg ver]

	interp alias {} pjs! {} variable pjs
	pjs!
	set pjs [dict create]
}

proc setupGui {} {
	package require central::extras
	package require gub
	namespace import ::gub::*
	package require meb
	namespace import ::meb::meb
	package require tu::toplevels::tk
	namespace import ::tu::toplevels::tk::*
	package require tu::epm
	namespace import ::tu::epm::*
	package require tu::pretty
	namespace import ::tu::pretty::hexdump
	package require Tk
	package require autoscroll
	::aitcs::aitcs new gui
	createGui
}

#---------------------------------------------------------------------
# Go
#---------------------------------------------------------------------

proc go {} {
	package require Tcl 8.6
	package require thatch
	namespace import ::thatch::*
	package require aitcs
	namespace import ::aitcs::aitcs
	package require tu
	package require tu::file
	namespace import ::tu::*
	namespace import ::tu::file::*
	setup
	cfg {*}[args {*}$::argv]; set ::argv {}; set ::argc 0
#	if {[info exists [cfg file]]} { set cfg($cfg(what)file) $cfg(file) }
	if {[cfg usage]}         { puts [usage]; return }
	if {[cfg help]}          { puts [usage]\n[help]; if {[cfg verbosity] > 1} { puts \nREADME:\n[readme] }; if {[cfg verbosity] > 2} { puts \nLICENSE:\n[join [dict values [LICENSE]] \n]}; return }
	if {[cfg printpreamble]} { puts [preamble] }
	if {[cfg printboot]}     { puts [string map [list @VER@ [package provide thatch]] $::thatch::bootstrap] }
	if {[cfg printpreamble] || [cfg printboot]} { return }
	autoloadConfig
	if {[cfg gui]} {
		setupGui
		autoloadProjects
		if {[cfg inFile] ne ""} {
			if {[file extension [cfg inFile]] eq ".tp"} {
				projectOpen [cfg inFile]
			} else {
				openThatch [cfg inFile]
			}
		}
	} else {
		cgogo
	}
}

namespace current}]::go

# EOF
